/*******************************************************************************
 *	BdsTestCreateData.cc	Creatse test data files in different formats
 *				T.Barnaby,	BEAM Ltd,	2009-11-17
 *******************************************************************************
 *
 * This program creates, water-marked, seismic data files in various formats
 * for test purposes.
 */
#include <iostream>
#include <iomanip>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <BdsD.h>
#include <BdsC.h>
#include <BFile.h>
#include <BdsLib.h>
#include <sys/stat.h>

#include <gcf2.h>

using namespace std;
using namespace Bds;

class Config {
public:
				Config();
	int			verbose;
	BString			network;
	BString			array;
	BString			source;
	BList<SelectionChannel>	channels;
	BTimeStamp		startTime;
	BTimeStamp		endTime;
	BString			format;
	BString			fileName;
	int			errMissingBlocks;
};

Config::Config(){
	verbose = 0;
	errMissingBlocks = 0;
}

class TestDataFile {
public:
				TestDataFile(Config& config);
	virtual			~TestDataFile();

	virtual BError		open() = 0;
	virtual void		close() = 0;
	virtual BError		writeHeader() = 0;
	virtual BError		writeData(DataBlock& dataBlock) = 0;
	BUInt			numChannels();
	BUInt			numSamples();
	double			sampleRate();
protected:
	Config&			oconfig;
	BFile			ofile;
	BUInt			onumChannels;
	BUInt			onumSamples;
	double			osampleRate;
};

TestDataFile::TestDataFile(Config& config) : oconfig(config){
}

TestDataFile::~TestDataFile(){
}

BUInt TestDataFile::numChannels(){
	return onumChannels;
}

BUInt TestDataFile::numSamples(){
	return onumSamples;
}

double TestDataFile::sampleRate(){
	return osampleRate;
}
	
/*******************************************************************************
 *	Bdrs data format
 *******************************************************************************
 */
class TestDataFileBdrs : public TestDataFile {
public:
				TestDataFileBdrs(Config& config);
	BError		open();
	void		close();
	BError		writeHeader();
	BError		writeData(DataBlock& dataBlock);
private:
};

TestDataFileBdrs::TestDataFileBdrs(Config& config) : TestDataFile(config){
	if(config.format == "BDRS-MM"){
		onumChannels = 38;
		onumSamples = 100;
		osampleRate = 20.0;
	}
	else {
		onumChannels = 20;
		onumSamples = 100;
		osampleRate = 20.0;
	}
}

BError TestDataFileBdrs::open(){
	BString	fileName;
	
	fileName.printf("input/%04d%03d_%02d%02d%02d.bdrs", oconfig.startTime.year(), oconfig.startTime.yday() + 1,
		oconfig.startTime.hour(), oconfig.startTime.minute(), oconfig.startTime.second());
	return ofile.open(fileName, "w");
}

void TestDataFileBdrs::close(){
	ofile.close();
}

BError TestDataFileBdrs::writeHeader(){
	BError			err;
	
	return err;
}

BError TestDataFileBdrs::writeData(DataBlock& dataBlock){
	BError			err;
	BUInt8			header[6];
	BUInt8			footer[6];
	BInt16			data[100][20];
	BUInt			c;
	BUInt			s;
	
//	printf("writeData\n");
	header[0] = 0x80 | 'T';
	header[1] = ((dataBlock.startTime.year() % 10) << 4) | ((dataBlock.startTime.yday() + 1) / 100);
	header[2] = ((((dataBlock.startTime.yday() + 1) % 100) / 10) << 4) | ((dataBlock.startTime.yday() + 1) % 10);
	header[3] = ((dataBlock.startTime.hour() / 10) << 4) | (dataBlock.startTime.hour() % 10);
	header[4] = ((dataBlock.startTime.minute() / 10) << 4) | (dataBlock.startTime.minute() % 10);
	header[5] = ((dataBlock.startTime.second() / 10) << 4) | (dataBlock.startTime.second() % 10);

	footer[0] = 0;
	footer[1] = header[0];
	footer[2] = header[0];
	footer[3] = header[0];
	footer[4] = header[0];
	footer[5] = header[0];

	
	if(oconfig.format == "BDRS-MM"){
		for(s = 0; s < onumSamples; s++){
			for(c = 0; c < 18; c++){
				data[s][c] = (BInt16)dataBlock.channelData[c][s];
			}
		}
		for(s = 0; s < onumSamples/10; s++){
			for(c = 18; c < 28; c++){
				data[(s * 10) + (c - 18)][18] = (BInt16)(0.5 * pow(2.0,15.0) * sin(2.0 * M_PI * s / 10));
			}
			for(c = 28; c < 38; c++){
				data[(s * 10) + (c - 28)][19] = (BInt16)(0.5 * pow(2.0,15.0) * cos(2.0 * M_PI * s / 10));
			}
		}
	}
	else {
		for(s = 0; s < onumSamples; s++){
			for(c = 0; c < onumChannels; c++){
				data[s][c] = (BInt16)dataBlock.channelData[c][s];
			}
		}
	}

	ofile.write(header, sizeof(header));
	ofile.write(data, sizeof(data));
	ofile.write(footer, sizeof(footer));

	return err;
}




BError waterMark(Config& oconfig, DataBlock& dataBlock){
	BError	err;
	BString	mark;
	BInt64	v;
	double*	d = 0;
	int	n;
	double	checkSum = 0;
	BUInt	c;
	BUInt	s;

	for(c = 0; c < dataBlock.channelData.size(); c++){
		checkSum = 0;
		for(s = 0; s < dataBlock.channelData[c].size(); s++){
			checkSum += double((BInt64)dataBlock.channelData[c][s] & ~(0xFFULL));
		}

		mark = "WMARK";
		mark = mark + "," + oconfig.network;
		mark = mark + "," + oconfig.array;
		mark = mark + "," + dataBlock.startTime.getString();
		mark = mark + "," + dataBlock.endTime.getString();

		if(oconfig.channels.number() > 0){
			mark = mark + "," + oconfig.channels[c].station;	// Station
			mark = mark + "," + oconfig.channels[c].channel;	// Channel
			mark = mark + "," + oconfig.channels[c].source;		// Source
		}
		else {
			mark = mark + "," + "";				// Station
			mark = mark + "," + "";				// Channel
			mark = mark + "," + "";				// Source
		}
		mark = mark + "," + BString::convert(dataBlock.channelData[c].size());	// BlockSize
		mark = mark + "," + BString::convert(checkSum, 1);	// Checksum

		if(mark.len() > int(dataBlock.channelData[c].size()))
			return err.set(1, "WaterMark to big for number of samples in a block");

//		printf("Mark: %d %s\n", mark.len(), mark.retStr());

		d = &dataBlock.channelData[c][0];
		for(n = 0; n < mark.len() + 1; n++){
			v = (((BInt64)d[n]) & ~(0xFFULL)) | mark[n];
			d[n] = v;
		}
	}
	return err;
}

BError createData(Config& config){
	BError			err;
	BFile			file;
	BUInt			blockNumber = 0;
	BUInt			numBlocks = 100;
	DataBlock		dataBlock;
	BTimeStamp		ts = config.startTime;
	BUInt			c;
	BUInt			s;
	double			pos = 0;
	double			freq;
	TestDataFile*		dataFile;
	
	// Determine format info
	if(config.format == "BDRS"){
		dataFile = new TestDataFileBdrs(config);
	}
	else if(config.format == "BDRS-MM"){
		dataFile = new TestDataFileBdrs(config);
	}
	else {
		return err.set(1, BString("Unknown format: ") + config.format);
	}
	
	// Calculate parameters
	numBlocks = round(1e-6 * BTimeStamp::difference(config.endTime, config.startTime) * dataFile->sampleRate() / dataFile->numSamples());
	freq = 1.0 * 2.0 * M_PI / dataFile->sampleRate();

	if(config.verbose){
		printf("NumBlocks: %d\n", numBlocks);
		printf("Create Data File: %s\n", config.fileName.retStr());
	}

	if(err = dataFile->open())
		return err;

	if(dataFile->writeHeader())
		return err;
	
	dataBlock.channelData.resize(dataFile->numChannels());
	for(c = 0; c < dataFile->numChannels(); c++){
		dataBlock.channelData[c].resize(dataFile->numSamples());
	}
	
	for(blockNumber = 0; blockNumber < numBlocks; blockNumber++){
		dataBlock.startTime = ts;
		dataBlock.endTime = dataBlock.startTime;
		dataBlock.endTime.addSeconds(5);
		dataBlock.channelNumber = 1;
		dataBlock.segmentNumber = 1;

		if(!config.errMissingBlocks || ((blockNumber % 10) != 9)){
			for(s = 0; s < dataFile->numSamples(); s++){
				for(c = 0; c < dataBlock.channelData.size(); c++){
					dataBlock.channelData[c][s] = 0.9 * pow(2.0,15.0) * sin((freq * pos) - (c * (M_PI / dataBlock.channelData.size())));
				}
				pos++;
			}

			if(err = waterMark(config, dataBlock))
				return err;
			if(err = dataFile->writeData(dataBlock))
				return err;
		}
		
		ts = dataBlock.endTime;
	}
	
	dataFile->close();

	return err;
}


void usage(void) {
	cerr << "Usage:\bdsTestCreateData [options] <fileName>\n";
	cerr << " -help              - Help on command line parameters\n";
	cerr << " -verbose           - Verbose print out\n";
	cerr << " -format <format>    - The data format\n";
	cerr << " -network <network> - The Network\n";
	cerr << " -source <source>   - The Source\n";
	cerr << " -startTime         - The StartTime\n";
	cerr << " -endTime           - The End Time\n";
	cerr << " -channels <station1:chan1,station2:chan2> - The channel information\n";
	cerr << "\n";
	cerr << " -errMissingBlocks  - Add missing blocks\n";
}

static struct option options[] = {
		{ "?",			0, NULL, 0 },
		{ "h",			0, NULL, 0 },
		{ "help",		0, NULL, 0 },
		{ "verbose",		0, NULL, 0 },
		{ "format",		1, NULL, 0 },
		{ "network",		1, NULL, 0 },
		{ "source",		1, NULL, 0 },
		{ "startTime",		1, NULL, 0 },
		{ "endTime",		1, NULL, 0 },
		{ "channels",		1, NULL, 0 },
		{ "errMissingBlocks",	0, NULL, 0 },
		{ 0,0,0,0 }
};

int main(int argc, char** argv){
	BError			err;
	int			optIndex = 0;
	int			c;
	BString			s;
	Config			config;
	BStringList		channels;
	BIter			i;
	
	// Set stdout to line buffering
	setlinebuf(stdout);

	config.network = "TT";
	config.array = "TSA";
	
	while((c = getopt_long_only(argc, argv, "", options, &optIndex)) == 0){
		s = options[optIndex].name;
		if(s == "help" || s == "h" || s == "?"){
			usage();
			return 1;
		}
		else if((s == "verbose") || (s == "v")){
			config.verbose = 1;
		}
		else if(s == "format"){
			config.format = optarg;
		}
		else if(s == "network"){
			config.network = optarg;
		}
		else if(s == "source"){
			config.source = optarg;
		}
		else if(s == "startTime"){
			config.startTime.setString(optarg);
		}
		else if(s == "endTime"){
			config.endTime.setString(optarg);
		}
		else if(s == "channels"){
			channels = bstringToList(optarg);
		}
		else if(s == "errMissingBlocks"){
			config.errMissingBlocks = 1;
		}
		else {
			usage();
			return 1;
		}
	}
	
	if((argc - optind) != 1){
		usage();
		return 1;
	}
	
	for(channels.start(i); !channels.isEnd(i); channels.next(i)){
		SelectionChannel	s;
		
		s.network = config.network;
		s.station = channels[i].pullToken(":");
		s.channel = channels[i].pullToken(":");
		s.source = config.source;
		
		config.channels.append(s);
	}
	
	config.fileName = argv[optind];

	if(!err){
		err = createData(config);
	}
	
	if(err)
		cerr << "Error: " << err.getString() << "\n";

	return err.getErrorNo();
}
