/*******************************************************************************
 *	BdsTestImportDirectData.cc	Create and import live test data
 *				T.Barnaby,	BEAM Ltd,	2010-03-08
 *******************************************************************************
 *
 * This program creates, water-marked, seismic data and directly imports it into
 * the BDS system.
 */
#include <iostream>
#include <iomanip>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <BdsD.h>
#include <BdsC.h>
#include <BFile.h>
#include <BdsLib.h>
#include <sys/stat.h>

#include <gcf2.h>

using namespace std;
using namespace Bds;

class Config {
public:
				Config();
	int			verbose;
	BString			network;
	BString			array;
	BString			source;
	BList<SelectionChannel>	channels;
	BTimeStamp		startTime;
	BTimeStamp		endTime;
	BString			format;
	BString			fileName;
	int			errMissingBlocks;
	BUInt			errMissingBlocksChannel;
};

Config::Config(){
	verbose = 0;
	errMissingBlocks = 0;
	errMissingBlocksChannel = 0;
}



BError waterMark(Config& oconfig, BUInt channel, DataBlock& dataBlock){
	BError	err;
	BString	mark;
	BInt64	v;
	double*	d = 0;
	int	n;
	double	checkSum = 0;
	BUInt	c;
	BUInt	s;

	for(c = 0; c < dataBlock.channelData.size(); c++){
		checkSum = 0;
		for(s = 0; s < dataBlock.channelData[c].size(); s++){
			checkSum += double((BInt64)dataBlock.channelData[c][s] & ~(0xFFULL));
		}

		mark = "WMARK";
		mark = mark + "," + oconfig.network;
		mark = mark + "," + oconfig.array;
		mark = mark + "," + dataBlock.startTime.getString();
		mark = mark + "," + dataBlock.endTime.getString();

		if(oconfig.channels.number() > 0){
			mark = mark + "," + oconfig.channels[channel + c].station;	// Station
			mark = mark + "," + oconfig.channels[channel + c].channel;	// Channel
			mark = mark + "," + oconfig.channels[channel + c].source;		// Source
		}
		else {
			mark = mark + "," + "";				// Station
			mark = mark + "," + "";				// Channel
			mark = mark + "," + "";				// Source
		}
		mark = mark + "," + BString::convert(dataBlock.channelData[c].size());	// BlockSize
		mark = mark + "," + BString::convert(checkSum, 1);	// Checksum

		if((mark.len() + 1) > int(dataBlock.channelData[c].size()))
			return err.set(127, BString("WaterMark to big for number of samples in a block: ") + mark.len());

//		printf("Mark: %d %s\n", mark.len(), mark.retStr());

		d = &dataBlock.channelData[c][0];
		for(n = 0; n < mark.len() + 1; n++){
			v = (((BInt64)d[n]) & ~(0xFFULL)) | mark[n];
			d[n] = v;
		}
	}
	return err;
}

BError createDataSM(DataAddAccess& bds, Config& config){
	BError			err;
	DataInfo		dataInfo;
	BIter			i;
	BUInt			numChannels = config.channels.size();
	BUInt			numBlocks = 100;
	BUInt			numSamples = numBlocks;
	double			sampleRate = 20.0;
	BUInt			blockNumber = 0;
	DataHandle		dataHandle;
	DataBlock		dataBlock;

	BTimeStamp		ts = config.startTime;
	BUInt			c;
	BUInt			s;
	double			pos = 0;
	double			freq;

	// Setup DataInfo
	dataInfo.startTime = config.startTime;
	dataInfo.endTime = config.endTime;
	dataInfo.array = config.array;
	dataInfo.description = "Test Data";
	dataInfo.synchronous = 1;
	dataInfo.channels.resize(numChannels);
	
	for(config.channels.start(i), c = 0; !config.channels.isEnd(i); config.channels.next(i), c++){
		DataChannel	dc;
		
		dc.startTime = config.startTime;
		dc.endTime = config.endTime;
		dc.network = config.channels[i].network;
		dc.station = config.channels[i].station;
		dc.channel = config.channels[i].channel;
		dc.source = config.channels[i].source;
		dc.dataFileChannel = 1 + c;
		dc.numBlocks = 0;
		dc.numSamples = 0;
		dc.sampleRate = sampleRate;
		dc.sampleFormat = SampleFormatFloat32;
		dc.importFormat = "TEST-SM";
		dc.importFilename = "PROGRAM";
		
		dataInfo.channels[c].append(dc);
	}

	// Setting up data stream
	if(err = bds.dataOpen(dataInfo, "w", "API-SM", 0, dataHandle))
		return err;

	// Setup info on this set of files
	if(err = bds.dataSetInfo(dataHandle, dataInfo))
		return err;

	freq = 1.0 * 2.0 * M_PI / sampleRate;

	ts = dataInfo.startTime;
	numSamples = 100;
	pos = 0;

	dataBlock.channelNumber = 1;
	dataBlock.segmentNumber = 0;
	dataBlock.channelData.resize(dataInfo.channels.size());
	for(c = 0; c < dataInfo.channels.size(); c++){
		dataBlock.channelData[c].resize(numSamples);
	}

	for(blockNumber = 0; ; blockNumber++){
		dataBlock.startTime = ts;
		dataBlock.endTime = dataBlock.startTime;
		dataBlock.endTime.addMicroSeconds(1e6 * numSamples / sampleRate);

		if(!config.errMissingBlocks || ((blockNumber % 10) != 9)){
			for(s = 0; s < numSamples; s++, pos++){
				for(c = 0; c < numChannels; c++){
					dataBlock.channelData[c][s] = 0.9 * pow(2.0,15.0) * sin(freq * pos);
				}
			}

			if(err = waterMark(config, 0, dataBlock)){
				if(config.verbose)
					cerr << "Warning: " << err.getString() << "\n";
				err.set(0);
			}

			if(err = bds.dataPutBlock(dataHandle, dataBlock))
				return err;
		}

		ts = dataBlock.endTime;
		if(ts > dataInfo.endTime)
			break;
	}

	err = bds.dataClose(dataHandle, err, 1);

	return err;
}

BError createDataCM(DataAddAccess& bds, Config& config){
	BError			err;
	DataInfo		dataInfo;
	BIter			i;
	BUInt			numChannels = config.channels.size();
	BUInt			numBlocks = 100;
	BUInt			numSamples = numBlocks;
	double			sampleRate = 20.0;
	BUInt			blockNumber = 0;
	DataHandle		dataHandle;
	DataBlock		dataBlock;

	BTimeStamp		ts = config.startTime;
	BUInt			c;
	BUInt			s;
	double			pos = 0;
	double			freq;

	// Setup DataInfo
	dataInfo.startTime = config.startTime;
	dataInfo.endTime = config.endTime;
	dataInfo.array = config.array;
	dataInfo.description = "Test Data";
	dataInfo.synchronous = 1;
	dataInfo.channels.resize(numChannels);
	
	for(config.channels.start(i), c = 0; !config.channels.isEnd(i); config.channels.next(i), c++){
		DataChannel	dc;
		
		dc.startTime = config.startTime;
		dc.endTime = config.endTime;
		dc.network = config.channels[i].network;
		dc.station = config.channels[i].station;
		dc.channel = config.channels[i].channel;
		dc.source = config.channels[i].source;
		dc.dataFileChannel = 1 + c;
		dc.numBlocks = 0;
		dc.numSamples = 0;
		dc.sampleRate = sampleRate;
		dc.sampleFormat = SampleFormatInt32;
		dc.importFormat = "TEST-CM";
		dc.importFilename = "PROGRAM";
		
		dataInfo.channels[c].append(dc);
	}

	// Setting up data stream
	if(err = bds.dataOpen(dataInfo, "w", "API-CM", 0, dataHandle))
		return err;

	// Setup info on this set of files
	if(err = bds.dataSetInfo(dataHandle, dataInfo))
		return err;

	freq = 1.0 * 2.0 * M_PI / sampleRate;

	for(c = 0; c < numChannels; c++){
		ts = dataInfo.startTime;
		numSamples = 200 + c;
		pos = 0;

		for(blockNumber = 0; ; blockNumber++){
			dataBlock.startTime = ts;
			dataBlock.endTime = dataBlock.startTime;
			dataBlock.endTime.addMicroSeconds(1e6 * numSamples / sampleRate);
			if(dataBlock.endTime > dataInfo.endTime){
				dataBlock.endTime = dataInfo.endTime;
				numSamples = round(1e-6 * BTimeStamp::difference(dataBlock.endTime, dataBlock.startTime) * sampleRate);
			}
			
			dataBlock.channelNumber = 1 + c;
			dataBlock.segmentNumber = 0;
			dataBlock.channelData.resize(1);
			dataBlock.channelData[0].resize(numSamples);

			if(!config.errMissingBlocks || ((c + 1) != config.errMissingBlocksChannel) || ((blockNumber % 10) != 9)){
				for(s = 0; s < numSamples; s++, pos++){
					dataBlock.channelData[0][s] = 0.9 * pow(2.0,15.0) * sin(freq * pos);
				}

				if(err = waterMark(config, c, dataBlock)){
					if(config.verbose)
						cerr << "Warning: " << err.getString() << "\n";
					err.set(0);
				}
#ifdef ZAP
				if(config.verbose)
					printf("Chan: %d Block: %d %s to %s %d\n", c, blockNumber, dataBlock.startTime.getString().retStr(), dataBlock.endTime.getString().retStr(), numSamples);
#endif
				if(err = bds.dataPutBlock(dataHandle, dataBlock))
					return err;
			}

			ts = dataBlock.endTime;
			if(ts >= dataInfo.endTime)
				break;
		}
	}

	err = bds.dataClose(dataHandle, err, 1);

	return err;
}


void usage(void) {
	cerr << "Usage: bdsTestImportData [options] <fileName>\n";
	cerr << " -help              - Help on command line parameters\n";
	cerr << " -verbose           - Verbose print out\n";
	cerr << " -host <hostname>   - BDS Server host name\n";
	cerr << " -user <user:password> - user id and password\n";
	cerr << " -format <format>   - The type of data (SM, CM)\n";
	cerr << " -array <array>     - The Array\n";
	cerr << " -network <network> - The Network\n";
	cerr << " -source <source>   - The Source\n";
	cerr << " -startTime         - The StartTime\n";
	cerr << " -endTime           - The End Time\n";
	cerr << " -channels <station1:chan1,station2:chan2> - The channel information\n";
	cerr << "\n";
	cerr << " -errMissingBlocks  - Add missing blocks\n";
	cerr << " -errMissingBlocksChannel <channel> - The channel to have missing blocks (all channels otherwise)\n";
}

static struct option options[] = {
		{ "?",			0, NULL, 0 },
		{ "h",			0, NULL, 0 },
		{ "help",		0, NULL, 0 },
		{ "verbose",		0, NULL, 0 },
		{ "host",		1, NULL, 0 },
		{ "user",		1, NULL, 0 },
		{ "format",		1, NULL, 0 },
		{ "array",		1, NULL, 0 },
		{ "network",		1, NULL, 0 },
		{ "source",		1, NULL, 0 },
		{ "startTime",		1, NULL, 0 },
		{ "endTime",		1, NULL, 0 },
		{ "channels",		1, NULL, 0 },
		{ "errMissingBlocks",	0, NULL, 0 },
		{ "errMissingBlocksChannel",	1, NULL, 0 },
		{ 0,0,0,0 }
};

int main(int argc, char** argv){
	BError			err;
	int			optIndex = 0;
	int			c;
	BString			s;
	Config			config;
	BStringList		channels;
	DataAddAccess		bds;
	BString			hostName;
	BString			user;
	BString			password;
	BIter			i;
	
	// Set stdout to line buffering
	setlinebuf(stdout);

	config.network = "TT";
	config.array = "TSA";
	
	while((c = getopt_long_only(argc, argv, "", options, &optIndex)) == 0){
		s = options[optIndex].name;
		if(s == "help" || s == "h" || s == "?"){
			usage();
			return 1;
		}
		else if((s == "verbose") || (s == "v")){
			config.verbose = 1;
		}
		else if(s == "host"){
			hostName = optarg;
		}
		else if(s == "user"){
			s = optarg;
			user = s.pullToken(":");
			password = s;
		}
		else if(s == "format"){
			config.format = optarg;
		}
		else if(s == "array"){
			config.array = optarg;
		}
		else if(s == "network"){
			config.network = optarg;
		}
		else if(s == "source"){
			config.source = optarg;
		}
		else if(s == "startTime"){
			config.startTime.setString(optarg);
		}
		else if(s == "endTime"){
			config.endTime.setString(optarg);
		}
		else if(s == "channels"){
			channels = bstringToList(optarg);
		}
		else if(s == "errMissingBlocks"){
			config.errMissingBlocks = 1;
		}
		else if(s == "errMissingBlocksChannel"){
			config.errMissingBlocksChannel = atoi(optarg);
		}
		else {
			printf("Unknown option: %s\n", s.retStr());
			usage();
			return 1;
		}
	}

	if((argc - optind) != 0){
		usage();
		return 1;
	}
	
	// Open BDS access
	err = bds.connectService(BString("//") + hostName + "/bdsDataAddAccess");

	// Connect to service
	if(!err)
		err = bds.connect(user, password);

	for(channels.start(i); !channels.isEnd(i); channels.next(i)){
		SelectionChannel	s;
		
		s.network = config.network;
		s.station = channels[i].pullToken(":");
		s.channel = channels[i].pullToken(":");
		s.source = config.source;
		
		config.channels.append(s);
	}
	
	config.fileName = argv[optind];

	if(!err){
		if(config.format == "SM")
			err = createDataSM(bds, config);
		else if(config.format == "CM")
			err = createDataCM(bds, config);
		else
			err.set(1, BString("Unknown format: ") + config.format);
	}
	
	if(err)
		cerr << "Error: " << err.getString() << "\n";

	return err.getErrorNo();
}
