/*******************************************************************************
 *	BdsTestValidateData.cc	Validate test data and meta data on BdsServer
 *				T.Barnaby,	BEAM Ltd,	2009-11-23
 *******************************************************************************
 *
 * This program validates the test data and MetaData in the BDS system.
 */
#include <iostream>
#include <iomanip>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <BdsD.h>
#include <BdsC.h>
#include <BFile.h>
#include <BdsLib.h>
#include <sys/stat.h>

using namespace std;
using namespace Bds;

class ChannelData {
public:
				ChannelData() : markStartPos(0), lastBlockPos(0){}
	BTimeStamp		startTime;
	BArray<BFloat64>	data;
	BUInt			markStartPos;
	BTimeStamp		lastBlockTime;
	BUInt			lastBlockPos;
};

class Config {
public:
				Config();
	int			verbose;
	int			sampleMultiplexed;
	Selection		selection;
	BString			fileName;
	BArray<ChannelData>	channelData;
};

Config::Config(){
	verbose = 0;
	sampleMultiplexed = 0;
}

BError waterMarkRead(DataBlock& dataBlock, BUInt channel, BDictString& info){
	BError		err;
	BString		mark;
	double*		d = &dataBlock.channelData[channel][0];
	int		n;
	char		ch;
	BList<BString>	list;
	
	for(n = 0; n < 1024; n++){
//		printf("Data: %llu\n", (BInt64)d[n]);
		ch = ((BInt64)d[n]) & (0xFFULL);
		if(ch == 0)
			break;
		mark += ch;
	}
	
	if(n >= 1024)
		return err.set(1, BString("No water mark found in dataBlock: ") + dataBlock.startTime.getString());

	list = mark.getTokenList(',');
	
//	printf("Mark: %s\n", mark.retStr());
	if(list.size() != 10)
		return err.set(1, BString("No valid water mark in dataBlock: ") + dataBlock.startTime.getString() + " list size: " + list.size());

	if(list[0] != "WMARK")
		return err.set(1, BString("No valid water mark in dataBlock: ") + dataBlock.startTime.getString() + " list size: " + list.size());

	info["magic"] = list[0];
	info["network"] = list[1];
	info["array"] = list[2];
	info["startTime"] = list[3];
	info["endTime"] = list[4];
	info["station"] = list[5];
	info["channel"] = list[6];
	info["source"] = list[7];
	info["blockSize"] = list[8];
	info["checksum"] = list[9];
	
	return err;
}

BError testValidateDataBlock(Config& config, BUInt channel, BUInt blockNumber, DataBlock& data){
	BError			err;
	BUInt			c;
	BUInt			s;
	BIter			i;
	double			checkSum;
	BDictString		info;
	BString			ci;

	for(c = 0; c < data.channelData.size(); c++){
		if(data.channelData[c].size() > 95){
			// Test water mark
			if(err = waterMarkRead(data, c, info)){
				// Ignore if there is not enough data
				if(err.getErrorNo() == 2)
					err.set(0);
				return err;
			}

			if(config.verbose > 1){
				printf("ChannelBase: %u Channel: %u Block: %u\n", channel, c, blockNumber);
				for(info.start(i); !info.isEnd(i); info.next(i)){
					printf("\t%s: %s\n", info.key(i).retStr(), info[i].retStr());
				}
			}
			ci.printf("ChannelBase: %u Channel: %u Block: %u", channel, c, blockNumber);

			checkSum = 0;
			for(s = 0; s < data.channelData[c].size(); s++){
				checkSum += double((BInt64)data.channelData[c][s] & ~(0xFFULL));
			}

			// Check data
			if(info["network"] != config.selection.channels[channel + c].network)
				return err.set(1, BString("Network incorrect: ") + ci + " " + info["network"] + " was " + config.selection.channels[channel + c].network);
			if(info["station"] != config.selection.channels[channel + c].station)
				return err.set(1, BString("Station incorrect: ") + ci + " " + info["station"] + " was " + config.selection.channels[channel + c].station);
			if(info["channel"] != config.selection.channels[channel + c].channel)
				return err.set(1, BString("Channel incorrect: ") + ci + " " + info["channel"] + " was " + config.selection.channels[channel + c].channel);
			if(info["source"] != config.selection.channels[channel + c].source)
				return err.set(1, BString("Source incorrect: ") + ci + " " + info["source"] + " was" + config.selection.channels[channel + c].source);
			if(info["startTime"] != data.startTime.getString())
				return err.set(1, BString("StartTime incorrect: ") + ci + " " + info["startTime"] + " was " + data.startTime.getString());
			if(info["endTime"] != data.endTime.getString())
				return err.set(1, BString("EndTime incorrect: ") + ci + " " + info["endTime"] + " was " + data.endTime.getString());
			if(info["checksum"] != BString::convert(checkSum, 1))
				return err.set(1, BString("CheckSum incorrect: ") + ci + " " + BString::convert(checkSum, 1) + " was " + info["checksum"]);
		}
	}
	
	return err;
}

BError waterMarkRead(BArray<BFloat64>& data, BUInt channel, BDictString& info, BUInt& startPos){
	BError		err;
	BString		mark;
	BUInt		n;
	char		ch;
	BList<BString>	list;
	int		startFound = 0;
	int		endFound = 0;
	
	for(n = 0; (n < data.size()) && !endFound; n++){
//		printf("Data: %llu\n", (BInt64)d[n]);
		ch = ((BInt64)data[n]) & (0xFFULL);
		if(startFound == 5){
			if(ch == 0)
				endFound = 1;
			mark += ch;
		}
		else if((startFound == 0) && (ch == 'W')){
			startFound++;
		}
		else if((startFound == 1) && (ch == 'M')){
			startFound++;
		}
		else if((startFound == 2) && (ch == 'A')){
			startFound++;
		}
		else if((startFound == 3) && (ch == 'R')){
			startFound++;
		}
		else if((startFound == 4) && (ch == 'K')){
			mark += "WMARK";
			startFound++;
			startPos = n - 4;
		}
	}
	
	if(!endFound && (data.size() > 1024))
		return err.set(1, BString("No water mark found in data"));

	if(!endFound)
		return err.set(127, BString("No water mark found in data"));


	list = mark.getTokenList(',');
	
//	printf("Mark: %s\n", mark.retStr());
	if(list.size() != 10)
		return err.set(1, BString("No valid water mark in data") + " list size: " + list.size());

	if(list[0] != "WMARK")
		return err.set(1, BString("No valid water mark in data: ") + " list size: " + list.size());

	if(list[8].retInt() > int(data.size() - startPos))
		return err.set(127, BString("Not enough data for full block yet"));

	info["magic"] = list[0];
	info["network"] = list[1];
	info["array"] = list[2];
	info["startTime"] = list[3];
	info["endTime"] = list[4];
	info["station"] = list[5];
	info["channel"] = list[6];
	info["source"] = list[7];
	info["blockSize"] = list[8];
	info["checksum"] = list[9];
	
	return err;
}

BError testValidateDataBlockStream(Config& config, BUInt channel, BUInt blockNumber, DataBlock& data, double sampleRate){
	BError			err;
	BUInt			c;
	BUInt			s;
	BIter			i;
	double			checkSum;
	BDictString		info;
	BString			ci;
	BUInt			startPos;
	BTimeStamp		startTime;

	// Add data to streams
	config.channelData.resize(channel + data.channelData.size());

	for(c = channel; c < data.channelData.size(); c++){
		if(config.channelData[c].startTime.year() == 0)
			config.channelData[c].startTime = data.startTime;
		config.channelData[c].lastBlockTime = data.startTime;
		config.channelData[c].lastBlockPos = config.channelData[c].data.size();
		config.channelData[c].data.append(data.channelData[c]);
	}

	// Validate channels
	for(c = 0; c < data.channelData.size(); c++){
// printf("Next Block Chan: %d\n", c);
		// Test water mark
		if(err = waterMarkRead(config.channelData[channel + c].data, c, info, startPos)){
			// Ignore if there is not enough data
			if(err.getErrorNo() == 127)
				err.set(0);
			return err;
		}

// printf("TestBlock at: %d\n", startPos);
		if(config.verbose > 1){
			printf("ChannelBase: %u Channel: %u Block: %u\n", channel, c, blockNumber);
			for(info.start(i); !info.isEnd(i); info.next(i)){
				printf("\t%s: %s\n", info.key(i).retStr(), info[i].retStr());
			}
		}
		ci.printf("ChannelBase: %u Channel: %u Block: %u", channel, c, blockNumber);

		// Work out what the blocks start time should be
		startTime = config.channelData[channel + c].startTime;
		startTime.addMicroSeconds(1e6 * startPos / sampleRate);

// printf("Checksum: Chan: %d Start: %d Num: %d %d\n", channel, startPos, info["blockSize"].retInt(), config.channelData[c].data.size());
		checkSum = 0;
		for(s = startPos; s < startPos + info["blockSize"].retInt(); s++){
			checkSum += double((BInt64)config.channelData[channel + c].data[s] & ~(0xFFULL));
		}

		// Check data
		if(info["network"] != config.selection.channels[channel + c].network)
			return err.set(1, BString("Network incorrect: ") + ci + " " + info["network"] + " was " + config.selection.channels[channel + c].network);
		if(info["station"] != config.selection.channels[channel + c].station)
			return err.set(1, BString("Station incorrect: ") + ci + " " + info["station"] + " was" + config.selection.channels[channel + c].station);
		if(info["channel"] != config.selection.channels[channel + c].channel)
			return err.set(1, BString("Channel incorrect: ") + ci + " " + info["channel"] + " was " + config.selection.channels[channel + c].channel);
		if(info["source"] != config.selection.channels[channel + c].source)
			return err.set(1, BString("Source incorrect: ") + ci + " " + info["source"] + " was " + config.selection.channels[channel + c].source);
		if(info["startTime"] != startTime.getString())
			return err.set(1, BString("StartTime incorrect: ") + ci + " " + info["startTime"] + " was" + startTime.getString());
		if(info["checksum"] != BString::convert(checkSum, 1))
			return err.set(1, BString("CheckSum incorrect: ") + ci + " " + BString::convert(checkSum, 1) + " was" + info["checksum"]);

		// Move to next block
//		printf("Move to block end: %d\n", config.channelData[c].lastBlockPos);
		if(config.channelData[channel + c].lastBlockPos == 0){
			config.channelData[channel + c].data.resize(0);
			config.channelData[channel + c].startTime.set(0, 1, 1);
			config.channelData[channel + c].lastBlockTime = data.startTime;
			config.channelData[channel + c].lastBlockPos = 0;
		}
		else {
			config.channelData[channel + c].data.del(0, config.channelData[channel + c].lastBlockPos);
			config.channelData[channel + c].startTime = config.channelData[channel + c].lastBlockTime;
			config.channelData[channel + c].lastBlockTime = data.startTime;
			config.channelData[channel + c].lastBlockPos = 0;
		}
	}
	
	return err;
}

BError testValidateData(DataAccess& bds, Config& config){
	BError			err;
	DataInfo		dataInfo;
	DataHandle		dataHandle;
	DataBlock		data;
	BUInt			c;
	BUInt			s;
	BUInt			b = 0;
	BIter			i;
	BTimeStamp		lastEndTime;

	if(config.verbose){
		printf("Validate Data\n");

		for(c = 0; c < config.selection.channels.number(); c++){
			printf("Chan: %d %s\n", c, bdsSelectionChannelInfo(config.selection, c).retStr());
		}
	}

	if(err = bds.dataSearch(config.selection, dataInfo))
		return err;
	
	if(!dataInfo.channels.size()){
		return err.set(1, "No Channels selected: Requires Network, Station/Array and Channels to be set to channels that exist");
	}
	
//	if(err = bds.dataOpen(dataInfo, "r", "API", 0, dataHandle)){
	if(err = bds.dataOpen(dataInfo, "r", "API", DataFlagClipDataToTime, dataHandle)){
		return err;
	}

	if(err = bds.dataGetInfo(dataHandle, 0, dataInfo)){
		return err;
	}

	if(config.sampleMultiplexed){
		for(s = 0; s < dataInfo.channels[0].size(); s++){
			while(1){
//				printf("Loop: Block: %u\n", blockNumber);
				if(err = bds.dataGetBlock(dataHandle, 0, s + 1, b, data)){
					if(err.getErrorNo() == ErrorEndOfData)
						err.set(0, "");
					return err;
				}

				if(data.endTime >= dataInfo.endTime)
					break;

				if(lastEndTime.isSet() && (lastEndTime != data.startTime))
					return err.set(1, BString("Block timestamp error: expecting: ") + lastEndTime.getString().retStr() + " got: " + data.startTime.getString().retStr());

				lastEndTime = data.endTime;

				if(err = testValidateDataBlockStream(config, 0, b, data, dataInfo.channels[0][0].sampleRate))
					return err;

				b++;
			}
		}
	}
	else {
		for(c = 0; c < dataInfo.channels.size(); c++){
			for(s = 0; s < dataInfo.channels[c].size(); s++){
				if(config.verbose)
					printf("Channel: %u Segment: %u NumBlocks: %u\n", c + 1, s + 1, dataInfo.channels[c][s].numBlocks);

				lastEndTime.clear();
				for(b = 0; b < dataInfo.channels[c][s].numBlocks; b++){
//					printf("Loop: Block: %u\n", b);
					if(err = bds.dataGetBlock(dataHandle, c + 1, s + 1, b, data)){
						return err;
					}

					if(lastEndTime.isSet() && (lastEndTime != data.startTime))
						return err.set(1, BString("Block timestamp error: expecting: ") + lastEndTime.getString().retStr() + " got: " + data.startTime.getString().retStr());

					lastEndTime = data.endTime;

					if(err = testValidateDataBlockStream(config, c, b, data, dataInfo.channels[c][s].sampleRate))
//					if(err = testValidateDataBlock(config, c, b, data))
						return err;
				}
			}
		}
	}
	bds.dataClose(dataHandle, BError(), 0);

	return err;
}



void usage(void) {
	cerr << "Usage:\bdsTestAuto [options]\n";
	cerr << " -help              - Help on command line parameters\n";
	cerr << " -host <hostname>   - BDS Server host name\n";
	cerr << " -user <user:password> - user id and password\n";
	cerr << " -verbose           - verbose\n";
	cerr << " -startTime         - The StartTime\n";
	cerr << " -endTime           - The End Time\n";
	cerr << " -select <network:station:channel:source>	- Select a data set. Can have multiple -select's\n";
	cerr << " -sampleMultiplexed - Access the data in a sample multiplexed form\n";
}

static struct option options[] = {
		{ "?",			0, NULL, 0 },
		{ "h",			0, NULL, 0 },
		{ "help",		0, NULL, 0 },
		{ "host",		1, NULL, 0 },
		{ "user",		1, NULL, 0 },
		{ "verbose",		0, NULL, 0 },
		{ "startTime",		1, NULL, 0 },
		{ "endTime",		1, NULL, 0 },
		{ "select",		1, NULL, 0 },
		{ "sampleMultiplexed",	0, NULL, 0 },
		{ 0,0,0,0 }
};

int main(int argc, char** argv){
	BError			err;
	int			optIndex = 0;
	int			c;
	BString			s;
	DataAccess		bds;
	BString			hostName;
	BString			user = "test";
	BString			password = "beam00";
	Config			config;
	BString			s1,s2,s3,s4;

	// Set stdout to line buffering
	setlinebuf(stdout);

	hostName = getenv("BDS_HOST");
	if(hostName == "")
		hostName = "localhost";

	while((c = getopt_long_only(argc, argv, "", options, &optIndex)) == 0){
		s = options[optIndex].name;
		if(s == "help" || s == "h" || s == "?"){
			usage();
			return 1;
		}
		else if(s == "host"){
			hostName = optarg;
		}
		else if(s == "user"){
			s = optarg;
			user = s.pullToken(":");
			password = s;
		}
		else if(s == "verbose"){
			config.verbose++;
		}
		else if(s == "sampleMultiplexed"){
			config.sampleMultiplexed = 1;
		}
		else if(s == "startTime"){
			if(err = config.selection.startTime.setString(optarg)){
				cerr << "Error: StartTime format error\n";
				usage();
				return 1;
			}
		}
		else if(s == "endTime"){
			if(err = config.selection.endTime.setString(optarg)){
				cerr << "Error: EndTime format error\n";
				usage();
				return 1;
			}
		}
		else if(s == "select"){
			s = optarg;
			s1 = s.pullToken(":");
			s2 = s.pullToken(":");
			s3 = s.pullToken(":");
			s4 = s.pullToken(":");
			config.selection.channels.append(SelectionChannel(s1, s2, s3, s4));
		}
		else {
			usage();
			return 1;
		}
	}
	
	if((argc - optind) != 0){
		usage();
		return 1;
	}

	// Open BDS access
	err = bds.connectService(BString("//") + hostName + "/bdsDataAccess");

	// Connect to service
	if(!err)
		err = bds.connect(user, password);

	if(!err){
		err = testValidateData(bds, config);
	}

	if(err)
		cerr << "Error: " << err.getString() << "\n";

	return err.getErrorNo();
}
